""" domible/src/domible/elements/scripting.py
"""

from typing import Any

from domible.elements.baseElements import BaseElement


class Canvas(BaseElement):
    """
    create a <canvas> element
    All I know of the <canvas> element is it's a problem for accessibility.
    As I learn more, I might build out this class to enforce as much accessibility as possible in a <canvas>.
    """

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="canvas", **kwArgs)


class NoScript(BaseElement):
    """
    create a <noscript> element
    I guess this is like a <div> that can contain general HTML
    to be rendered if the rowser doesn't support the desired scripting.
    """

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="noscript", **kwArgs)


class Script(BaseElement):
    """
    create a <script> element
    domible doesn't know anything about scripts.
    This class exists for the domible user to add scripts to their HTML.
    The content should be a string of valid text (probably JS) given the attributes of the element.
    """

    def __init__(self, contents: str, **kwArgs):
        super().__init__(tag="script", **kwArgs)


class Template(BaseElement):
    """
    create a <template> element
    I guess this is like a <div> that can contain general HTML
    to be rendered later using JavaScript
    """

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="template", **kwArgs)


## end of file



""" domible/src/domible/starterDocuments/barebones.py """

from domible.elements import Html, Head, Meta, Title, Body


def basicHeadEmptyBody(title: str = "Domible Default Title", lang: str = "en") -> Html:
    """
    barebones creates an Html object with the minimal recommended elements:
    DOCTYPE, head with two meta elements and a tile, and a body with nothing
    """

    return Html(
        lang=lang,
        contents=[
            Head(
                [
                    Meta(charset="UTF-8"),
                    Meta(**{"http-equiv": "X-UA-Compatible"}, content="IE=edge"),
                    Meta(
                        name="viewport", content="width=device-width, initial-scale=1.0"
                    ),
                    Title(title),
                ]
            ),
            Body(),
        ],
    )


# end of file



"""  domible/src/domible/elements/tables.py
it's all about HTML tables in here
And HTML tables are ONLY for presenting tabular data.
If you use any of these table related elements to format the layout of your document,
you should be hunted down and ... well, something unpleasant ought to happen to you 
considering the audible pain you've inflicted on screen reader users 
it's 2022, use  CSS for layout and formatting  
"""

from typing import Any

from .baseElements import BaseElement


#######
class TableData(BaseElement):
    """ create a td element """

    def __init__(self, contents: Any, **kwArgs):
        super().__init__(tag='td', contents=contents, **kwArgs)


#######
class TableHeader(BaseElement):
    """ create a th element """

    def __init__(self, contents: Any, **kwArgs):
        super().__init__('th', contents, **kwArgs)


#######
class TableColumnHeader(TableHeader):
    """ creates a th element with attribute scope="col" added by default.
    This is most likely used for th elements in the first row of the table.
    By setting the scope attribute to col, 
    a screen reader will read out the column names as the user explores a row. 
    """
    def __init__(self, contents: Any, **kwArgs):
        super().__init__(contents, scope='col', **kwArgs)


#######
class TableRowHeader(TableHeader):
    """ creates a th element with attribute scope="row" added by default.
    This is most likely used for th elements in the first column of the table.
    By setting the scope attribute to row, 
    a screen reader will read out the row names as the user explores a column. 
    """
    def __init__(self, contents: Any, **kwArgs):
        super().__init__(contents, scope='row', **kwArgs)


#######
class TableRow(BaseElement):
    """ creates a tr element.
    Note the tr element contains a list of th and/or td elements.
    The th/td elements can be provided at creation time,
    or added later via instance methods to add content.
    """
    def __init__(self, cells: list[Any] = None, **kwArgs):
        """
        cells should be a list of th or td elements 
        """
        super().__init__('tr', cells, **kwArgs)


"""
classes TableHead, TableBody, and TableFoot differ in only the four letters after the 't' for the tag string
hopefully there are no cut'n'paste errors here...
"""
#######


class TableHead(BaseElement):
    """ creates a thead element.
    Note the thead element contains a list of tr elements.
    The tr elements can be provided at creation time,
    or added later via instance methods to add content.
    """
    def __init__(self, rows: list[Any] = None, **kwArgs):
        """
        rows should be a list of tr elements 
        """
        super().__init__(tag='thead', contents=rows, **kwArgs)


#######
class TableBody(BaseElement):
    """ creates a tbody element.
    Note the tbody element contains a list of tr elements.
    The tr elements can be provided at creation time,
    or added later via instance methods to add content.
    """
    def __init__(self, rows: list[Any] = None, **kwArgs):
        """
        rows should be a list of tr elements 
        """
        super().__init__(tag='tbody', contents=rows, **kwArgs)


#######
class TableFoot(BaseElement):
    """ creates a tfoot element.
    Note the tfoot element contains a list of tr elements.
    The tr elements can be provided at creation time,
    or added later via instance methods to add content.
    """
    def __init__(self, rows: list[Any] = None, **kwArgs):
        """
        rows should be a list of tr elements 
        """
        super().__init__(tag='tfoot', contents=rows, **kwArgs)


#######
class Caption(BaseElement):
    """ creates a caption element. """
    def __init__(self, contents: Any, **kwArgs):
        if not contents:
            raise ValueError("a Caption must have some contents")
        super().__init__(tag='caption', contents=contents, **kwArgs)


#######
class Table(BaseElement):
    """ creates the table element."""
    def __init__(self, caption: Caption, contents: Any = None, **kwArgs):
        """
        a caption is not required in a table, here's where I start to get opinionated.
        And this opinion is toward accessibility.
        Tables with captions are nice when using single letter navigation.
        If I type 't' to go to the next table, I generally have to arrow up to understand what is in the table.
        A caption howerver will be read to me after typing 't' and landing on the next table.
        """
        if not caption or not isinstance(caption, Caption):
            raise ValueError(
                "Table caption needs to be of type domible.elements.Caption")
        self.caption = caption
        # make sure caption is first on the contents list so it gets rendered by the base __repr__
        contents = [caption] + contents if contents else [caption]
        super().__init__(tag='table', contents=contents, **kwArgs)


## end of file


""" domible/src/domible/builders/listBuilder.py

Provide a more flexible interface to the various list elements.

This builder was motivated during implementation of the NavBuilder.
It might be necessary to update attributes in an <li> element,
to, for example, indicate that list item is the current page in a <nav> element.
It's better to have functionality like that in a builder,
not in the element class itself.
"""

import logging
logger = logging.getLogger(__name__)

from typing import Any

from domible.elements import DescriptionTerm, DescriptionDef, DListItem, ListItem 
from domible.elements import OrderedList, UnorderedList, MenuList, DescriptionList
from domible.elements import Script, Template
from domible.utils import indexMatchingAttributes 


class ListBuilder:
    """ 
    used to create an HTML list element,
    e.g., <ul>, <ol>, <menu>.
    <dl> is a different builder due to its contents being structurally different.

    The type of list created is left unspecified until the HTML element is requested.
    This works because the allowed contents of each  type of list is the same, 
    the semantics of the list element are determined only by the tag string 
    (hopefully someone will correct me if I'm wrong about this).
    Does this provide any real value (the delaying of the tag specification)?
    Perhaps ther's a need to generate a <ul> and <ol> from the same set of list entries?
    Doubtful.
    It's six of one, half dozen the other though, when to specify the type of list, 
    so let's go with the delayed style.

    I toyed with the idea of a class ListEntry builder.
    It was motivated by the fact a list can contain <li>, <script>, and <template> elements.
    The ListEntry was going to, sort of, encapsulate that, 
    ideally making it easyer to create ListEntries.
    Then users of ListBuilder would not have to explicitly create an <li>, <script>, or <template> element before adding it to the ListBuilder object.
    ListEntry would hold the contents of the eventual HTML element, it would also have to hold attributes to eventually create the element. 
    ListEntry started getting messy and duplicating BaseElement functionality.
    I experimented with ways to avoid that duplication.
    Short answer, it was all worse than simply requiring the ListBuilder user to
    use ListItem, Script, or Template objects when creating/useing ListBuilder objects.

    I was also going down a rabbit hole deciding how much functionality to add to ListBuilder.
    I thought about all the cool stuff to do, but realized quickly it would be just as easy 
    for a user to grab the items from the ListBuilder and operate directly on that Python list.
    OOP authorities might legitimately have me thrown in the stocks for the idea of allowing 
    users of ListBuilder to operate directly on its data.
    I'll rationalize this anti pattern by claiming pragmatism 
    and let's see first if anyone will use this before going deep on its interface.

    Another thought on the ListBuilder interface, and why this class exists.
    self.items is guaranteed to be a list based on __init__,
    and Python's list interface is already quite good.
    self.items contents though are guaranteed to be a type of BaseElement.
    Considering that, the ListBuilder interface should probably provide functionality to optimize dealing with a list of BaseElement objects.
    For example, I might want to add or update attributes to all elements in the list.
    Or maybe I want to update an attribute in the contents of a list item.
    And here we go down the rabbit hole... 
    """
    def __init__(self, items: list[Any] = None, **attributes):
        """
        as noted in the comments on the class declaration, 
        the type of list, the tag string, 
        is not needed until the builder generates an HTML element.
        We do however want to be able to set attributes for the list, thus **attributes
        """
        self.items: list[Any] = items if items else list()
        if not isinstance(self.items, list): self.items = [self.items]
        self.attributes: dict[str, str] = attributes if attributes else dict()
        for item in self.items:
            if not isinstance(item, (ListItem, Script, Template)):
                raise TypeError(f"{item} -- is not an li, script, or template element")


    def addItem(self, item: Any, index: int = None, attributes: dict[str, str] = None, before: bool = True) -> None:
        """
        add an item to the list, where is based on values of other parameters. 
        if incdex given, add item at that index 
        if index is not in range, raise IndexError
        if attributes, add item based on the attributes of another item
        if before is True, ad new item before existing item with matching attributes, else after
        before is ignored if attributes is None 
        if attributes is specified but no match is found, raise ValueError 
        if no index and attributes, add new item to end of list
        if index and attributes are both specified, attributes is ignored.
        """
        if not isinstance(item, (ListItem, Script, Template)):
            raise TypeError(f"{item} -- is not an li, script, or template element")
        if index:
            if index >= 0 and index < len(self.items):
                self.items.insert(index, item)
            else:
                raise IndexError(f"index {index} is outside range, len of items is {len(self.items)}")
        elif attributes:
            if (insert := indexMatchingAttributes(self.items, attributes)) != None:
                print(f"inserting at index {insert}, before is {before}")
                if before: self.items.insert(insert, item)
                else: self.items.insert(insert + 1, item)
            else:
                raise ValueError(f"no element found matching attributes {attributes}")
        else:
            self.items.append(item)


    def getList(self, tag: str, **attributes):
        """ 
        return the HTML element.
        More attributes can be added to the lists' opening tag if desired. 
        """
        if attributes: self.attributes.update(attributes)
        ## self.items = [ListItem(item) if not isinstance(item, (ListItem, Script, Template)) else item for item in self.items]
        if tag == "ul": return UnorderedList(self.items, **self.attributes)
        elif tag == "ol": return OrderedList(self.items, **self.attributes)
        elif tag == "menu": return MenuList(self.items, **self.attributes)
        else: raise ValueError(f"cannot build list with tag {tag}")


## end of file 


""" domible/src/domible/elements/sectioning.py 
"""

from typing import Any
from domible.elements.baseElements import BaseElement


class Article(BaseElement):
    """create <article> element"""

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="article", contents=contents, **kwArgs)


class Body(BaseElement):
    """create <body> element"""

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="body", contents=contents, **kwArgs)


class Footer(BaseElement):
    """create <footer> element"""

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="footer", contents=contents, **kwArgs)


class Header(BaseElement):
    """create <header> element"""

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="header", contents=contents, **kwArgs)


class Heading(BaseElement):
    """create heading (<h1, ..., h6>) element."""

    def __init__(self, level: int, contents: Any = None, **kwArgs):
        self.level = level
        super().__init__(tag=f"h{level}", contents=contents, **kwArgs)


class Main(BaseElement):
    """create <main> element"""

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="main", contents=contents, **kwArgs)


class Nav(BaseElement):
    """
    create a <nav> element
    For accessibility, require a label which will be used to set an aria-label attribute.
    This will result in a meaningful announcement when navigating by landmarks (and other scenarios). 
    """

    def __init__(self, label: str, contents: Any = None, **kwArgs):
        self.label = label
        super().__init__(tag="nav", contents=contents, **{"aria-label": self.label}, **kwArgs)


class Section(BaseElement):
    """create <section> element"""

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="section", contents=contents, **kwArgs)


## end of file



""" domible/src/domible/elements/__init__.py
"""

##
# only expose base classes for typing 
# for example, cases where a container can operate on any BaseClass derived element.
##
from domible.elements.baseElements import BaseElement, BaseVoidElement 

from domible.elements.roots import Html
from domible.elements.meta import Base, Head, Meta, Style, Title 
from domible.elements.inline import Anchor, Div, Paragraph, Span
from domible.elements.lists import DescriptionTerm, DescriptionDef, DListItem, ListItem, OrderedList, UnorderedList, MenuList, DescriptionList
from domible.elements.scripting import Canvas, NoScript, Script, Template 
from domible.elements.sectioning import Article, Body, Footer, Header, Heading, Main, Nav, Section 
from domible.elements.tables import TableData, TableHeader, TableColumnHeader, TableRowHeader, TableRow, TableHead, TableBody, TableFoot, Caption, Table

# end of file



""" domible/src/domible/__init__.py 
"""

# read version from installed package
from importlib.metadata import version
__version__ = version("domible")

import domible.elements
import domible.builders
import domible.starterDocuments 

## end of file 


""" domible/src/domible/builders/tableBuilder.py

The tableBuilder module is the first attempt at providing some level of abstraction to creating a more complicated HTML element. 
dataclasses are used to represent HTML table related elements using any datatypes. 
RowInfo and TableInfo classes have properties with semantics indicating their use within the table.
The specific HTML elements used to construct the table are encapsulated within the classes.  

The user can create the rows however they like then add them to the TableInfo object.

There is a correlation between the dict defining a row and dict defining the column headings in the TableInfo.
That correlation is in the keys of the dicts.
The dict defining the column headings defines which values within a row are included in the table.
Thus a row can be a dict with an arbitrary list of key, value pairs.
Then what is included in a table is defined by the column headings dict.
When a table is being generated from TableInfo,
the keys from the column headings dict is sent to each RowInfo.
The RowInfo object uses that to get the value for the cell for that column.
If the RowInfo entries dict has a matching key, the associated value is used for that cell (a <td> element.)
If there is no matching key, the cell will be empty (or use None) for that row for that column .
The RowInfo entries dict can have entries that are not included in a specific table.
That will be the case when the table column IDs are a subset of the keys in the RowInfo entries dict. 

The values in the column headings dict can be used to have custom names for the column headings. 
The default name of the column heading is the key itself, e.g.:
columnHeadings = {'col1':'col1', 'col2':'col2', ... }
to use better names for the column headings, try something like:
customColumnHeadings = {'col1':'Column One', 'col2':'Column Two', ... }
"""


import logging

logger = logging.getLogger(__name__)

from dataclasses import dataclass, field
from typing import Any, Dict

from domible.elements import (
    Caption,
    Table,
    TableRow,
    TableData,
    TableColumnHeader,
    TableRowHeader,
    TableHead,
    TableBody,
    Style,
    Script,
)


#######
@dataclass
class RowBuilder:
    """
    RowBuilder has the values for the cells of the row stored as a dict (with the row heading called out separately)
    the key in the entries dict is the column id, the value is what will eventually be in the <th> or <td> elements
    the first item is called out as the heading for that row mainly as an API feature.
    """

    heading: Any = "you need to set the row heading"
    # this is how **kwArgs works in dataclasses
    # to ensure each instantiation gets its own object,
    # not a reference to a dict shared by all RowInfo objects
    entries: Dict = field(default_factory=dict)

    #####
    def addEntries(self, **kwArgs) -> None:
        """
        add entries to the row by passing in any number of named parameters (columnId = 'some value').
        the Ids represent the columns,
        the value is the entry in that column for this row
        """
        if self.entries:
            self.entries.update(kwArgs)
        else:
            self.entries = kwArgs

    #####
    def properties(self) -> list[Any]:
        """
        The properties are the keys from the entries in this row.
        The keys are used to identify which values from a row are used in a particular table.
        """
        if self.entries:
            return list(self.entries.keys())
        else:
            return []

    #####
    def getRow(self, columns: list[Any]) -> TableRow:
        """
        return the TableRow object that can generate the HTML for the row
        the columns parameter is needed to know which values from entries to include in the row
        """
        rowCells = [TableRowHeader(self.heading)]
        rowCells += [TableData(self.entries.get(col)) for col in columns]
        """
        rowCells is now a list of Elements (representing <th> and <td> elements)
        it can be used to create a TableRow (representing a <tr> element) object 
        """
        return TableRow(rowCells)


#######
@dataclass
class TableBuilder:
    """
    an object of TableBuilder holds all the information needed to generate an HTML table
    """

    caption: Any
    # the 0th column, i.e., the column of row headings
    # the upper left cell, 0,0.
    # Hopefully this is set to something meaningful by the user of this class.
    rowHeadingName: Any = "Row Names"
    columnHeadings: Dict = field(default_factory=dict)
    rows: list[RowBuilder] = field(default_factory=list)

    #####
    def addRow(self, row: RowBuilder) -> None:
        self.rows.append(row)

    #####
    def generateColumnHeadings(self) -> None:
        """
        if you don't want to set the column headings directly, they can be inferred by the names of properties in the rows

        this method will iterate through the rows and keep track of the property name (key) of each item (key, value pair) in a row
        The columns can then be arranged based on the position of the propertys in the rows (the default),
        or another option can be specified (TODO)
        """
        counts = {}
        for row in self.rows:
            for prop in row.properties():
                if not counts.get(prop):
                    counts[prop] = 1
                else:
                    counts[prop] += 1
        # for now, column headings are all the properties found in the rows
        # TODO: allow the user to specify ordering of the columns based on counts
        #    or maybe columnHeadings.sort()
        # self.columnHeadings = list(counts.keys())
        self.columnHeadings = {key: key for key in counts.keys()}

    #####
    def getTable(self) -> tuple[Table, Style, Script]:
        """
        need to go through self (TableInfo) and convert all the info
        to corresponding elemensts (tr, th, td, caption...)
        """
        logger.debug(f"getting table with caption {self.caption}")
        table = Table(Caption(self.caption))
        # first add the column headings to the table
        # if column headings hve not been set yet, generate them based on the  rows
        if len(self.columnHeadings) == 0:
            self.generateColumnHeadings()
        columnNames = [self.rowHeadingName] + list(self.columnHeadings.values())
        logger.debug(
            f"table with caption {self.caption}, has column heading names:\n {columnNames}"
        )
        """
        split this to multiple lines for readability 
        first add the table head <thead> element to the table.
        a <thead> is a list of rows though in this case it's a single row of the names of the columns
        a TableRow is a list of anything derived from TableData <td> or TableHeader <th> elements 
        """
        table.addContent(
            TableHead(
                [
                    TableRow(
                        # list comprehension to initiate the TableRow
                        [TableColumnHeader(colName) for colName in columnNames]
                    )  # close of TableRow construction
                ]
            )  # close TableHead construction including close list for list of TableRows to construct TableHead
        )  # close of addContent
        """
        now add all the rows in the tbody element 
        same idea as adding the TableHead above 
        """
        table.addContent(
            TableBody(
                [row.getRow(list(self.columnHeadings.keys())) for row in self.rows]
            )
        )
        return table, None, None


#######
def buildTableFromDicts(caption: Any, rows: list[Dict]) -> Table:
    """
    this is, so far, the easiest way to construct an HTML table from data the user has pulled from anywhere
    The rows are the dicts in the list, in order they appear in the list
    the column names are derived from the keys in the dicts
    the row heading is the value from the first item in the dict representing that row
    and the heading for the first column (column of row headings) in the table is the key from the first key, value  pair from the first dict in the list

    Using this a developer doesn't need to know anything about generating HTML tables, especially accessibility
    it's all about defaults...
    Caveat, I've only used this for lists of very consistent dicts,
    that is, they all have the same size and keys.
    It's likely this fails quickly for random dicts in a list
    That is, if the dicts represent, say a very sparsely populated table, dicts with different keys with little overlap
    this table might not be what you want.
    It's better to use the TableInfo and RowInfo dataclasses
    """
    if len(rows) == 0:
        raise ValueError("no rows sent to createTableFromDict")
    rowHeadingName = list(rows[0].keys())[0]
    tbuilder = TableBuilder(caption, rowHeadingName)
    for row in rows:
        items = list(row.items())
        heading = items[0][1]
        entries = dict(items[1:])
        ri = RowBuilder(heading, entries)
        tbuilder.addRow(ri)
    return tbuilder.getTable()


# end of file



"""  domible/src/domible/elements/lists.py
Here we have elements specific to creating lists.
the list items (including <dt> and <dd>) are defined in here as well.
"""

from typing import Any

from domible.elements.baseElements import BaseElement

"""
<li>, <dt>, and <dd> are nearly the same element with different tags.
<dt> has more restrictions on its permitted content but I'm not sure this package needs to enforce that.
Maybe at some point I'll implement more types around content categories and build in more enforcement.
For now, I'm taking the C approach, so go ahead and corrupt memory.
Or in this case, implement nonsensical HTML documents.
"""


class DescriptionTerm(BaseElement):
    """creates a dt element."""

    def __init__(self, contents: Any, **kwArgs):
        super().__init__(tag="dt", contents=contents, **kwArgs)


class DescriptionDef(BaseElement):
    """creates a dd element."""

    def __init__(self, contents: Any, **kwArgs):
        super().__init__(tag="dd", contents=contents, **kwArgs)


class DListItem:
    """
    This is an attempt to make the dl list element behave similar to the other list elements.
    It is designed to be analogous to an li elment.
    It is created with a list of terms and their definition.
    It renders itself since it is not derived from BaseElement.
    Screen readers seem to have trouble with dl lists, they're a pain all around.
    """

    def __init__(self, terms: list[DescriptionTerm], definition: DescriptionDef):
        self.terms = terms
        self.definition = definition

    def __repr__(self):
        entryString = ""
        for term in self.terms:
            entryString += f"{term}"
        entryString += f"{self.definition}"
        return entryString


class ListItem(BaseElement):
    """creates an li element."""

    def __init__(self, contents: Any, **kwArgs):
        super().__init__(tag="li", contents=contents, **kwArgs)


class UnorderedList(BaseElement):
    """creates a ul element."""

    def __init__(self, entries: list[Any] = None, **kwArgs):
        super().__init__("ul", entries, **kwArgs)


class OrderedList(BaseElement):
    """creates an ol element."""

    def __init__(self, entries: list[Any] = None, **kwArgs):
        super().__init__("ol", entries, **kwArgs)


class MenuList(BaseElement):
    """creates a menu element."""

    def __init__(self, entries: list[Any] = None, **kwArgs):
        super().__init__("menu", entries, **kwArgs)


class DescriptionList(BaseElement):
    """creates a dl element."""

    def __init__(self, entries: list[Any] = None, **kwArgs):
        super().__init__("dl", entries, **kwArgs)


## end of file



""" domible/src/distarter/__init__.py 
"""

# read version from installed package
# using same version as domible as it's the main package 
from importlib.metadata import version
__version__ = version("domible")


from . import main

def run():
    main.run()

## end of file



""" domible/src/distarter/main.py 

this file serves as Python code to get started with domible
The datetime and webbrowser imports and openPage() funtion 
are not needed to use domible.
They're here only to be able to open the starter page in your default browser.
"""

from datetime import datetime as dt
import webbrowser as wb

from domible.elements import Html, Body, Title
from domible.elements import Heading, Anchor, Paragraph
from domible.starterDocuments import basicHeadEmptyBody


def openPage(htmlDoc: Html) -> None:
    """create temp html file to use webbrowser to open passed in Html doc"""
    thf = f"tmp_html_{dt.timestamp(dt.now())}.html"
    with open(thf, "w") as f:
        f.write(f"{htmlDoc}")
    wb.open(thf)


#######
def run() -> None:
    """
    generate, and show in browser,
    a very simple HTML document
    """
    title = "BareBones with a Minimal Body"
    bga = Anchor(
        href="https://blindgumption.com", contents="the Blind Gumption Website"
    )
    htmlDoc = basicHeadEmptyBody(title)
    body = htmlDoc.getBodyElement()
    body.addContent(
        [
            Heading(1, title),
            Paragraph(f"You might find more interesting content at {bga}"),
        ]
    )
    openPage(htmlDoc)


if __name__ == "__main__":
    run()

## end of file



""" domible/src/domible/elements/meta.py 

the meta tags, e.g., meta  :)
and others, e.g., title, link, base, style, script  
"""

from typing import Any

from .baseElements import BaseElement, BaseVoidElement


class Base(BaseVoidElement):
    """ generates the base voide element """

    def __init__(self, **kwArgs):
        """ base is a void element, no contents are allowed """
        super().__init__(tag="base", **kwArgs)


class Head(BaseElement):
    def __init__(self, contents: Any = None, **kwArgs):
        """ 
        tried having a constructor argument for a title with a default string.
        Turned in to a pain considering a Title object could be in the contents.
        Leaving it up to a document builder to handle a title string argument.  
        """
        super().__init__(tag="head", contents=contents, **kwArgs)


class Meta(BaseVoidElement):
    """ generates the meta voide element """

    def __init__(self, **kwArgs):
        """ meta is a void element, no contents are allowed """
        super().__init__(tag="meta", **kwArgs)


class Style(BaseElement):
    """ 
    create a <style> element 
    As it stands, domible doesn't know anything about CSS.
    The user crating a <style> object is expected to create valid CSS 
    and create this element with the CSS content as a string.
    """
    def __init__(self, contents: str, **kwArgs):
        super().__init__(tag="style", **kwArgs)


class Title(BaseElement):
    """ Create a Title element
    Titles can only contain strings of text (per spec).
    This is enforced in this class.

    it's necessary to override setContent and addContent to ensure title is a string 
    and there is no other content in the element 
    """

    def __init__(self, contents: str, **kwArgs):
        if not isinstance(contents, str):
            raise TypeError("Title contents must be a string")
        self.titleStr = contents
        super().__init__(tag='title', contents=self.titleStr, **kwArgs)


    def setTitle(self, title: str) -> None:
        """ set the value, must be a string, for the Title. """
        self.setContent(title)


    def setContent(self, content: Any) -> None:
        """ 
        <title> can only have a string that is the title  
        If the new content is a string, 
        overrite any existing title 
        else, raise an exception 
        """
        if not isinstance(content, str):
            raise TypeError("Title contents must be a string")
        self.titleStr = content
        super().setContent(content)


    def addContent(self, content: Any, front: bool = False) -> None:
        """ 
        let Title.setContent handle this 
        """
        self.setContent(content)


# end of file



"""  domible/src/domible/elements/inline.py 
"""
from typing import Dict, Any

from domible.elements.baseElements import BaseElement


class Anchor(BaseElement):
    def __init__(self, href: str, contents: Any = None, **kwArgs):
        self.href = href
        super().__init__(tag="a", contents=contents, href=self.href, **kwArgs)


class Div(BaseElement):
    """create a div element"""

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="div", contents=contents, **kwArgs)


#######
class Paragraph(BaseElement):
    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="p", contents=contents, **kwArgs)


#######
class Span(BaseElement):
    """create a span element
    use for wrapping a single content .
    For multiple elements, use a div.
    """

    def __init__(self, contents: Any = None, **kwArgs):
        super().__init__(tag="span", contents=contents, **kwArgs)


## end of file



""" domible/src/domible/tools.py

tools.py has functions for users of the domible package to do simple things
like open their HTML doc in the browser,
or save the HTML to a specified file.
tools.py imports other modules within domible thus should not be used by any of domible's submodules.
Any helper functions needed by domible should be put in utils.py 
See the top level comments in utils.py to understand why tools.py exists. 
"""

from pathlib import Path 
from tempfile import NamedTemporaryFile
import webbrowser as wb

from domible.elements import BaseElement 
from domible.elements import Html


def open_in_browser(htmlDoc: Html) -> None:
    """create temp file to use webbrowser to open passed in Html doc"""
    path =NamedTemporaryFile(delete=False, suffix='.html')
    f=open(path.name, 'w+t')
    f.write(f"{htmlDoc}")
    f.close()
    wb.open('file://' + path.name)


def save_to_file(element: BaseElement, filename: str, force: bool = False) -> None:
    """ 
    save the passed in element to passed in filename.
    If the file name exists and is a regular file, 
    save will fail unless force is Tru 
    """
    fp = Path(filename)
    if fp.exists():
        # if the file exists, is a regular file and force is true, carry on
        # otherwise, raise a FileExists error
        if not  fp.is_file() or not force:
            raise FileExistsError(f"{filename} exists and is not a regular file, or force is False")
    # if file does exists, force must be True
    with fp.open('w+t') as f:
        f.write(f"{element}")


## end of file



""" domible/src/domible/builders/__init__.py """

from domible.builders.tableBuilder import RowBuilder, TableBuilder, buildTableFromDicts
from domible.builders.navBuilder import NavBuilder 
from domible.builders.listBuilder import ListBuilder 


# end of file
